using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Runtime.InteropServices;
using GrayScaleClass;

namespace PDFDecNetPro
{
    public enum TPDF_Rejection_Reasons
    {
       PDF417_OK                   =   0,
       PDF417_Rectangle_Error      =  -1, // no Start and Stop Patterns were found
       PDF417_ColCount_Error       =  -4,
       PDF417_RowCount_Error       =  -5,
       PDF417_ECCodeword_Error     =  -6,
       PDF417_ErasuresLimit_Error  =  -7, // too many codewords with improper clasters
       PDF417_RS_Error             =  -8, // too many errors in Reed Solomon code
       PDF417_CharEncodation_Error =  -9, // unexpected encodation scheme
       PDF417_StartPattern_Error   = -10, // Stop pattern is not found
       PDF417_StopPattern_Error    = -11, // Stop pattern is not found
       PDF417_Indicators_Error     = -12, // incorrect data in row indicators
       PDF417_Geometry_Error       = -13, // unallowed module size
       PDF417_NOMEMORY             = -99,
       PDF417_UNKNOWN              = -100,
       PDF417_DISCONNECTED         = -200
    }
    
   public enum TPDF_Break_Reason
   {
        ALL_INSPECTED = 0,
        TIMEOUT = 1,
        TERMINATED = 2
   }
    
   public enum  TPDF417_SpeedMode
   {
       SP_Fast   = 1,
       SP_Robust = 0,
       SP_Smart  = 2
   }
    
   public enum TPDF417_DirectMode
   {
       DM_LeftRight  = 0,
       DM_Horizontal = 1,
       DM_All        = 2,
       DM_Vertical   = 3
   }

   public struct TPDF417_Quality
   {
      // 0..100                    , -1 = have not been calculated
      public float symbol_contrast;            // symbol contrast(SC)  = Rmax-Rmin
      public float min_reflectance;            // Rmin
      public float max_reflectance;            // Rmax
      public float global_threshold;           // global_threshold(GT) = (Rmax+Rmin)/2
      public float min_edge_contrast;          // ECmin
      public float modulation;                 // modulation (MOD)     = ECmin/SC
      public float defects;                    //
      public float decodability;               // V
      public float width_height_proportion;
      public float unused_error_correction;    // UEC

      // grades
      // 0..4,  0 = Worst , 4 = Best, -1 = have not been calculated
      public float decode_grade;
      public float symbol_contrast_grade;
      public float min_reflectance_grade;
      public float min_edge_contrast_grade;
      public float modulation_grade;
      public float defects_grade;
      public float decodability_grade;
      public float unused_error_correction_grade;
      public float overall_grade;
    }

   public class TPDF417_Decoder_Exception : Exception
   {
      public TPDF417_Decoder_Exception(String msg)
         : base(msg)
      {
      }
   }

   public class TPDF417_Decoder
   {
#pragma warning disable 649
      internal unsafe struct TPDF417_Info // //result of decoding per each PDF417 symbol
      {
         public Int32 RejectionReason;     // Result

         public Int32 rowcols0;    // symbol corners
         public Int32 rowcols1;
         public Int32 rowcols2;
         public Int32 rowcols3;
         public Int32 rowcols4;
         public Int32 rowcols5;
         public Int32 rowcols6;
         public Int32 rowcols7;

         public Int32 rcFlag;        // true if a qudrangle symbol was found
         public Int32 RSErrorCount;  // number of Reed Solomon errors
         public Int32 ColCount;      // horizontal dimension (from 1 to 30)
         public Int32 RowCount;      // vertical dimension   (from 4 to 90)
         public byte* RejectionString; // pointer to error string
         public Int32 ECLevel;       // Error Correction Level (from 0 to 8)
         public Int32 pchlen;        // length of decoded byte array
         public byte* pch;// decoded byte array
         public byte* c_str;       // decoded char array (points to the same allocation in memory)
         public TPDF417_Quality Quality;
      }
       
      internal unsafe struct TPDF417_OptMode
       {
          public Int32 MaxSymbolCount;          // maximum number of decoded PDF417 labels
          public Int32 SpeedMode;  //  SP_Fast,  SP_Robust
          public Int32 CalcGrade;  //  CG_No , CG_Yes ;
          public Int32 DirectMode; //  DM_LeftRight, DM_Horizontal, DM_All
          public Int32 BWFilter;                  // 0: none, +1 => BW+1,  -1 => BW-1
          public Int32 TimeOut;                   // in milliseconds, zero - not timeout
          public Int32 ScanStep;
       }

       internal unsafe struct TPDF417_ImageInfo
       {
          public Int32 SymbolCount;       // count of decoded PDF417 symbols
          public Int32 RejectionReason;   // Result
          public byte* RejectionString; // pointer to error string
          public Int32 BreakReason;       //
          public byte* BreakString;     // pointer to error string
       }
#pragma warning restore 649

       private const string LIB_NAME = "PDF_PRO32.dll";
       //private const string LIB_NAME = "PDF_PRO64.dll";

       [DllImport(LIB_NAME, EntryPoint = "Connect_PDF417_Decoder", CallingConvention = CallingConvention.StdCall)]
       internal static extern IntPtr Connect_PDF417_Decoder([In] Int32 MaxRowCount, [In] Int32 MaxColCount);

       [DllImport(LIB_NAME, EntryPoint = "Disconnect_PDF417_Decoder", CallingConvention = CallingConvention.StdCall)]
       internal static extern void Disconnect_PDF417_Decoder([In] IntPtr Decoder);

       [DllImport(LIB_NAME, EntryPoint = "Create_PDF417_Options", CallingConvention = CallingConvention.StdCall)]
       internal static extern IntPtr Create_PDF417_Options([In] IntPtr Decoder, [In]TPDF417_OptMode opt);

       [DllImport(LIB_NAME, EntryPoint = "Delete_PDF417_Options", CallingConvention = CallingConvention.StdCall)]
       internal static extern void Delete_PDF417_Options([In] IntPtr Options);

       [DllImport(LIB_NAME, EntryPoint = "GetPDF417_ImageInfo", CallingConvention = CallingConvention.StdCall)]
       internal static extern IntPtr GetPDF417_ImageInfo([In] IntPtr DecoderO);

       [DllImport(LIB_NAME, EntryPoint = "GetPDF417_Info", CallingConvention = CallingConvention.StdCall)]
       internal static extern IntPtr GetPDF417_Info([In] IntPtr DecoderO, [In] Int32 SymbolNumber);

       [DllImport(LIB_NAME, EntryPoint = "Decode_PDF417_Bits", CallingConvention = CallingConvention.StdCall)]
       internal static extern int Decode_PDF417_Bits([In] IntPtr DecoderO, [In] Int32 RowCount, [In] Int32 ColCount, [In] IntPtr[] PPBits);

       private IntPtr FDecoderW = IntPtr.Zero;
       private IntPtr DecoderO = IntPtr.Zero;
       private bool optChanged = true;
       private int pdfCount = 0;
       private TPDF_Rejection_Reasons rejectionReasons = TPDF_Rejection_Reasons.PDF417_UNKNOWN;

       private TPDF417_OptMode optmode = new TPDF417_OptMode();
       
       public TPDF417_Decoder(int MaxRowCount, int MaxColCount)
       {
          optmode.MaxSymbolCount = 1;
          optmode.SpeedMode = (Int32)TPDF417_SpeedMode.SP_Fast;
          optmode.CalcGrade = 0;
          optmode.DirectMode = (Int32)TPDF417_DirectMode.DM_LeftRight;
          optmode.BWFilter = 0; // -1
          optmode.TimeOut = 0;
          optmode.ScanStep = 4;

          FDecoderW = Connect_PDF417_Decoder(MaxRowCount, MaxColCount);
          if (FDecoderW == null)
             throw new TPDF417_Decoder_Exception("Connect_PDF417_Decoder return null!");
       }
       
       public void Disconnect ( )
        {
            unsafe
            {
               if (DecoderO != IntPtr.Zero)
               {
                  Delete_PDF417_Options(DecoderO);
                  DecoderO = IntPtr.Zero;
               }
               if (FDecoderW != IntPtr.Zero)
               {
                  Disconnect_PDF417_Decoder(FDecoderW);
                  FDecoderW = IntPtr.Zero;
               }
            }
        }

       public int DecodeImage(Bitmap BMP)
       {
          pdfCount = 0;
          rejectionReasons = TPDF_Rejection_Reasons.PDF417_UNKNOWN;
          GrayScale G = new GrayScale(BMP);
          G.OnDecBits += this.DecodeBitsO;

          G.DecodeBits();

          IntPtr info = GetPDF417_ImageInfo(DecoderO);
          unsafe
          {
             TPDF417_ImageInfo* UII = (TPDF417_ImageInfo*)info.ToPointer();
             {
                pdfCount = UII->SymbolCount;
                rejectionReasons = (TPDF_Rejection_Reasons)UII->RejectionReason;
             }
          }
          return pdfCount;
       }

       public String GetSymbolData(int codepage, int symbolIdx)
       {
          if (pdfCount == 0) return null;
          StringConvertor SC = new StringConvertor( codepage );
          
          IntPtr ptr = GetPDF417_Info(DecoderO, (Int32)symbolIdx);
          unsafe
          {
             TPDF417_Info* info = (TPDF417_Info*)ptr.ToPointer();
             byte[] resBuf = new byte[info->pchlen];
             for (int i = 0; i < info->pchlen; i++)
                resBuf[i] = info->pch[i];
             return SC.ByteArrayToString(resBuf);
          }
       }

       public int[] GetSymbolRect(int symbolIdx)
       {
          if (pdfCount == 0) return null;
          int[] res = new int[8];
          IntPtr ptr = GetPDF417_Info(DecoderO, (Int32)symbolIdx);
          unsafe
          {
             TPDF417_Info* info = (TPDF417_Info*)ptr.ToPointer();
             res[0] = info->rowcols0;
             res[1] = info->rowcols1;
             res[2] = info->rowcols2;
             res[3] = info->rowcols3;
             res[4] = info->rowcols4;
             res[5] = info->rowcols5;
             res[6] = info->rowcols6;
             res[7] = info->rowcols7;
          }
          return res;
       }

       public int GetSymbolRSErrorCount(int symbolIdx)
       {
          IntPtr ptr = GetPDF417_Info(DecoderO, (Int32)symbolIdx);
          unsafe
          {
             TPDF417_Info* info = (TPDF417_Info*)ptr.ToPointer();
             return info->RSErrorCount;
          }
       }

       public TPDF417_Quality GetSymbolQuality(int symbolIdx)
       {
          if (pdfCount == 0) throw new TPDF417_Decoder_Exception("No decoded!");
          TPDF417_Quality res = new TPDF417_Quality();
          IntPtr ptr = GetPDF417_Info(DecoderO, (Int32)symbolIdx);
          unsafe
          {
             TPDF417_Info* info = (TPDF417_Info*)ptr.ToPointer();
             res = info->Quality;
          }
          return res;
       }

       public int SymbolCount()
       {
          return pdfCount;
       }

       public TPDF_Rejection_Reasons RejectionReasons()
       {
          return rejectionReasons;
       }

       public void SetMaxSymbolCount(int count)
       {
          optmode.MaxSymbolCount = count;
          optChanged = true;
       }

       public void SetSpeedMode(TPDF417_SpeedMode mode)
       {
          optmode.SpeedMode = (Int32)mode;
          optChanged = true;
       }

       public void SetCalcGrade(Boolean calc)
       {
          if (calc)
             optmode.CalcGrade = 1;
          else
             optmode.CalcGrade = 0;
          optChanged = true;
       }

       public void SetDirectMode(TPDF417_DirectMode dir)
       {
          optmode.DirectMode = (Int32)dir;
          optChanged = true;
       }

       public void SetBWFilter(int filter)
       {
          optmode.BWFilter = filter;
          optChanged = true;
       }

       public void SetScanStep(int step)
       {
          optmode.ScanStep = step;
          optChanged = true;
       }

       public void SetQualityOn(bool calcOn)
       {
          if(calcOn)
             optmode.CalcGrade = 1;
          else
             optmode.CalcGrade = 0;
          optChanged = true;
       }

       public void SetTmeOut(int timeOut)
       {
          optmode.TimeOut = timeOut;
          optChanged = true;
       }

        private int DecodeBitsO (Int32 RowCount, Int32 ColCount, IntPtr[ ] PPBits)
        {
           if ((DecoderO == IntPtr.Zero) || (optChanged))
            {
              unsafe
              {
                 if (DecoderO != IntPtr.Zero)
                 {
                    fixed (void* PO = &DecoderO)
                    {
                       IntPtr P = new IntPtr(PO);
                       Delete_PDF417_Options(P);
                    }
                 }
                 DecoderO = Create_PDF417_Options(FDecoderW, optmode);
                 if (DecoderO == null)
                    throw new TPDF417_Decoder_Exception("Failed Create_PDF417_Options!");
              }
              optChanged = false;
            }
            return Decode_PDF417_Bits(DecoderO, RowCount, ColCount, PPBits);
        }
    }
}
